from quillsql.assets.pgtypes import PG_TYPES

def convert_type_to_postgres(data_type_id):
   # find the object in PG_TYPES that matches the type
    pg_type = next((pg_type for pg_type in PG_TYPES if pg_type['oid'] == data_type_id), None)
    return pg_type['typname'] if pg_type else data_type_id


import json

def remove_fields(query_result, cursor_description, fields_to_remove):
  fields = [
      {"name": desc[0], "dataTypeID": desc[1]}
      for desc in cursor_description
      if desc[0] in fields_to_remove
  ]
  rows = [dict(json.loads(json.dumps(row, default=str))) for row in query_result]

  for row in rows:
    for field in fields_to_remove:
      if field in row:
          del row[field]
  return {fields, rows}

def array_to_map(queries, array_to_map, metadata, target_pool):
  for i in range(len(queries)):
    query_result = target_pool.query(queries[i])
    metadata[array_to_map.get("arrayName")][i][array_to_map.get("field")] = query_result.get("rows")



from quillsql.db.postgres import format_postgres, connect_to_postgres, get_schema_column_info_postgres, get_tables_by_schema_postgres, run_query_postgres, disconnect_from_postgres
from quillsql.db.bigquery import format_bigquery_config, connect_to_bigquery, get_schema_column_info_big_query, get_tables_by_schema_big_query, run_query_big_query

def get_db_credentials(database_type, connection_string):
  if(database_type.lower() == 'postgresql'):
    return format_postgres(connection_string)
  elif(database_type.lower() == 'bigquery'):
    return format_bigquery_config(connection_string)
  return {}

def connect_to_db(database_type, config, using_connection_string):
    if(database_type.lower() == 'postgresql'):
        return connect_to_postgres(config, using_connection_string)
    elif(database_type.lower() == 'bigquery'):
        return connect_to_bigquery(config, using_connection_string)
    return None

def run_query_by_db(database_type, query, connection):
    if (database_type.lower() == 'postgresql'):
        return run_query_postgres(query, connection)
    elif (database_type.lower() == 'bigquery'):
        return run_query_big_query(query, connection)
    return None

def disconnect_from_db(database_type, connection):
    if (database_type.lower() == 'postgresql'):
        return disconnect_from_postgres(connection)
    return None

def get_schema_tables_by_db(database_type, connection, schema_name):
    if (database_type.lower() == 'postgresql'):
        return get_tables_by_schema_postgres(connection, schema_name)
    elif (database_type.lower() == 'bigquery'):
        return get_tables_by_schema_big_query(connection, schema_name)
    return None

def get_schema_column_info_by_db(database_type, connection, schema_name, table_names):
    if (database_type.lower() == 'postgresql'):
        return get_schema_column_info_postgres(connection, schema_name, table_names)
    elif (database_type.lower() == 'bigquery'):
        return get_schema_column_info_big_query(connection, schema_name, table_names)
    return None


import os
from dotenv import load_dotenv

import requests
import redis


from quillsql.db.cached_connection import CachedConnection
from quillsql.db.db_helper import get_db_credentials, get_schema_column_info_by_db, get_schema_tables_by_db
from quillsql.utils.schema_conversion import convert_type_to_postgres
from quillsql.utils.run_query_processes import (
    array_to_map,
    remove_fields,
)

load_dotenv()

ENV = os.getenv("PYTHON_ENV")
DEV_HOST = "http://localhost:8080"
PROD_HOST = "https://quill-344421.uc.r.appspot.com"
HOST = DEV_HOST if ENV == "development" else PROD_HOST


## Quill - Fullstack API Platform for Dashboards and Reporting.
class Quill:
    def __init__(
        self,
        private_key,
        database_type,
        database_connection_string=None,
        database_config=None,
        metadataServerURL=None,
        cache=None,
    ):
        # Handles both dsn-style connection strings (eg. "dbname=test password=secret" )
        # as well as url-style connection strings (eg. "postgres://foo@db.com")
        self.baseUrl = metadataServerURL if metadataServerURL != None else HOST
        if database_connection_string != None:
          self.target_connection = CachedConnection(database_type, get_db_credentials(database_type, database_connection_string), cache, True)
        else:
          self.target_connection = CachedConnection(database_type, database_config, cache, False)
        self.private_key = private_key

    def get_cache(self, cache_config):
        cache_type = cache_config and cache_config.get("cache_type")
        if cache_type and cache_type == "redis" or cache_type == "rediss":
            return redis.Redis(
                host=cache_config.get("host", "localhost"),
                port=cache_config.get("port", 6379),
                username=cache_config.get("username", "default"),
                password=cache_config.get("password"),
            )
        return None

    def query(self, org_id, data):
        metadata = data.get("metadata")
        if not metadata:
            return {"error": "400", "errorMessage": "Missing metadata."}

        task = metadata.get("task")
        if not task:
            return {"error": "400", "errorMessage": "Missing task."}

        try:
            pre_query_results = self.run_queries(
                metadata.get("preQueries"),
                self.target_connection.database_type,
                metadata.get('databaseType'),
                metadata,
                metadata.get("runQueryConfig"),
            )  # used by the view task to get non-sensitive data
            if metadata.get("runQueryConfig") and metadata.get("runQueryConfig").get('overridePost'):
                return {
                    "data": pre_query_results ,
                    "status": "success"
                }
            view_query = None
            if (metadata.get("preQueries")):
                view_query = metadata.get("preQueries")[0]
            payload = {
                **metadata,
                "orgId": org_id,
                "preQueryResults": pre_query_results,
                "viewQuery": view_query
            }
            quill_results = self.post_quill(metadata.get("task"), payload)
            if quill_results.get("error"):
                return {"error": quill_results.get("error"), "status": "error"}
            # If there is no metedata in the quill results, create one
            if not quill_results.get("metadata"):
                quill_results["metadata"] = {}
            metadata = quill_results.get("metadata")
            final_query_results = self.run_queries(
                quill_results.get("queries"),  self.target_connection.database_type, metadata.get('databaseType'), metadata, metadata.get("runQueryConfig")
            )
            # Quick hack to make the sdk work with the Frontend
            if len(final_query_results.get("queryResults")) == 1:
                query_result = final_query_results.get("queryResults")[0]
                quill_results["metadata"]["rows"] = query_result.get("rows")
                quill_results["metadata"]["fields"] = query_result.get("fields")
            return {
                "data": quill_results.get("metadata"),
                "queries": final_query_results,
                "status": "success",
            }

        except Exception as err:
            return {"error": str(err), "status": "error"}

    def run_queries(self, queries, pkDatabaseType, databaseType, metadata=None, runQueryConfig=None):
        results = {}
        if not queries:
            return {"queryResults": []}
        if databaseType and databaseType.lower() != pkDatabaseType.lower():
            return {"dbMismatched": True, "backendDatbaseType": pkDatabaseType}
        if runQueryConfig and runQueryConfig.get("arrayToMap"):
            array_to_map(
                queries, runQueryConfig.get("arrayToMap"), metadata, self.target_connection
            )
            return {"queryResults": []}
        elif runQueryConfig and runQueryConfig.get("getColumns"):
            query_results = self.target_connection.query(queries[0].strip().rstrip(";") + ' limit 1')
            results["columns"] = [{'name': result['name'], 'dataTypeID': convert_type_to_postgres(result['dataTypeID'])} for result in query_results['fields']]
        elif runQueryConfig and runQueryConfig.get("getTables"):
            tables = get_schema_tables_by_db(
              self.target_connection.database_type,
              self.target_connection.connection,
              runQueryConfig['schemaNames']
            )
            schema = get_schema_column_info_by_db(
              self.target_connection.database_type,
              self.target_connection.connection,
              runQueryConfig['schemaNames'],
              tables
            )
            results["queryResults"] = schema
        else:
            if runQueryConfig and runQueryConfig.get("limitThousand"):
                queries = [query.strip().rstrip(";") + " limit 1000" for query in queries]
            elif runQueryConfig and runQueryConfig.get("limitBy"):
                queries = [query.strip().rstrip(";") + " limit " + runQueryConfig.get("limitBy") for query in queries]
            query_results = [self.target_connection.query(query) for query in queries]
            results["queryResults"] = query_results
            if runQueryConfig and runQueryConfig.get("removeFields"):
                results["queryResults"] = remove_fields(
                    query_results, runQueryConfig.get("removeFields")
                )
            if runQueryConfig and runQueryConfig.get("convertDatatypes"):
                for query_result in results["queryResults"]:
                    query_result["fields"] = [
                        {
                            "name": field["name"],
                            "displayName": field["name"],
                            "field": field["name"],
                            "isVisible": True,
                            "dataTypeID": field["dataTypeID"],
                            "fieldType": convert_type_to_postgres(field["dataTypeID"]),
                        }
                        for field in query_result["fields"]
                    ]

        return results

    def post_quill(self, path, payload):
        url = f"{self.baseUrl}/sdk/{path}"
        headers = {"Authorization": f"Bearer {self.private_key}"}
        response = requests.post(url, json=payload, headers=headers)
        return response.json()



import psycopg2
from psycopg2.extensions import make_dsn
from quillsql.assets.pgtypes import PG_TYPES

def format_postgres(connection_string ):
    to_dsn = lambda conn: make_dsn(conn) if "://" in conn else conn
    return to_dsn(connection_string)

def connect_to_postgres(config, usingConnectionString):
    if usingConnectionString:
      return psycopg2.connect(config)
    else:
      return psycopg2.connect(
        database=config['dbname'],
        user=config['user'],
        password=config['password'],
        host=config['host'],
        port=config['port']
      )

def run_query_postgres(query, connection):
    cursor = connection.cursor()
    cursor.execute(query)
    result = cursor.fetchall()
    fields = [
            {"name": desc[0], "dataTypeID": desc[1]} for desc in cursor.description
        ]
    cursor.close()
    rows_dict = [dict(zip([field['name'] for field in fields], row)) for row in result]
    return {"rows": rows_dict, "fields": fields}

def disconnect_from_postgres(connection):
    connection.close()
    return

# getTablesBySchemaPostgres
def get_tables_by_schema_postgres(connection, schema_names):
    all_tables = []
    for schema_name in schema_names:
      query = f"SELECT table_name, table_schema FROM information_schema.tables WHERE table_schema = '{schema_name}'"
      results = run_query_postgres(query, connection)
      for row in results['rows']:
        cur_table = {}
        cur_table['table_name'] = row['table_name']
        cur_table['schema_name'] = row['table_schema']
        all_tables.append(cur_table)
    return all_tables

# getSchemaColumnInfoPostgress
def get_schema_column_info_postgres(connection, schema_name, table_names):
    all_columns = []
    for table_name in table_names:
        query = f"SELECT column_name, udt_name FROM information_schema.columns WHERE table_schema = '{table_name['schema_name']}' AND table_name = '{table_name['table_name']}' ORDER BY ordinal_position"
        results = run_query_postgres(query, connection)
        columns = []
        for row in results['rows']:
            # Convert row['udt_name'] to postgresql oid
            pg_type = next((pg_type for pg_type in PG_TYPES if pg_type['typname'] == row['udt_name']), None)
            if pg_type == None:
                pg_type = 1043
            columns.append({
                'columnName': row['column_name'],
                'displayName': row['column_name'],
                'dataTypeID': pg_type['oid'],
                'fieldType': row['udt_name'],
            })
        all_columns.append({
            'tableName': table_name['schema_name']+'.'+table_name['table_name'],
            'displayName': table_name['schema_name']+'.'+table_name['table_name'],
            'columns': columns
        })
    return all_columns


from setuptools import setup, find_packages

setup(
    name="quillsql",
    version="2.1.1",
    packages=find_packages(),
    install_requires=[
        "psycopg2-binary",
        "requests",
        "redis",
        "python-dotenv",
        "pytest",
    ],
    author="Quill",
    author_email="shawn@quillsql.com",
    description="Quill SDK for Python.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/quill-sql/quill-python",
)


from google.cloud import bigquery
from google.oauth2 import service_account
import json
import re

def format_bigquery_config(connection_string):
    # find the start of the json {
    json_start = connection_string.find('{')
    if (json_start == -1) :
      raise Exception("Invalid input string. No JSON data found.");
    
    dataset_name = connection_string[0:json_start].strip()
    json_string = connection_string[json_start:]
    try:
      service_account = json.loads(json_string)
      if (not service_account.get('project_id') or not service_account.get('private_key')) :
        raise Exception( "Invalid service account JSON. Required fields are missing.")
      
      return {
        "dataset_id": dataset_name,
        "project": service_account.get('project_id'),
        "credentials": service_account,
      }
    except:
       print('Invalid JSON string')
    return connection_string

def connect_to_bigquery(config, using_connection_string):
    if using_connection_string:
      credentials = service_account.Credentials.from_service_account_info(config['credentials'])
    else:
      credentials = service_account.Credentials.from_service_account_file(config['service_account_file_path'])
    return bigquery.Client(project=config['project'], credentials=credentials)

def run_query_big_query(query, connection):
    query_job = connection.query(query)
    result = query_job.result()
    rows = [dict(row) for row in result]
    fields = [{"name": field.name, "dataTypeID": convert_bigquery_to_postgres(field.field_type)} for field in result.schema]
    # TODO CONVERT to postgres types

    return {"rows": rows, "fields": fields}

def get_tables_by_schema_big_query(connection, schema_names):
    all_table = []
    for schema_name in schema_names:
      dataset_ref = connection.dataset(schema_name)
      tables = connection.list_tables(dataset_ref)
      for table in tables:
         cur_table = {}
         cur_table['table_name'] = table.table_id
         cur_table['schema_name'] = schema_name
         all_table.append(cur_table)
    return all_table

def get_schema_column_info_big_query(connection, schema_name, table_names):
    all_columns = []
    for table_name in table_names:
        table_ref = connection.dataset(table_name['schema_name']).table(table_name['table_name'])
        table = connection.get_table(table_ref)
        columns = []
        for field in table.schema:
            columns.append({
                'columnName': field.name,
                'displayName': field.name,
                'dataTypeId': convert_bigquery_to_postgres(field.field_type),
                'fieldType': field.field_type
            })
        all_columns.append({
            'tableName': table_name['schema_name']+'.'+table_name['table_name'],
            'displayName': table_name['schema_name']+'.'+table_name['table_name'],
            'columns': columns
        })
    return all_columns

def infer_schema_big_query(elem):
  # compare elem with regex
  if isinstance(elem, list):
     return 23
  if isinstance(elem, object):
    if re.match(r"/^\d{4}-\d{2}-\d{2}$/", elem.get('value')):
      return 1082
    elif re.match(r"/^\d{2}\/\d{2}\/\d{2,4}$/", elem.get('value')):
      return 1082
    elif re.match(r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?Z$/", elem.get('value')):
      return 1184
    elif re.match(r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?$/", elem.get('value')):
      return 1114
    elif re.match(r"/^\d{2}:\d{2}:\d{2}$/", elem.get('value')):
      return 1083
  if isinstance(elem, str):
    if re.match(r"/^\d{4}-\d{2}-\d{2}$/", elem):
      return 1082
    elif re.match(r"/^\d{2}\/\d{2}\/\d{2,4}$/", elem):
      return 1082
    elif re.match(r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?Z$/", elem):
      return 1184
    elif re.match(r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?$/", elem):
      return 1114
    elif re.match(r"/^\d{2}:\d{2}:\d{2}$/", elem):
      return 1083
    else:
      return 1043
  return 1043

def convert_bigquery_to_postgres(value):
  type_to_oid = {
    "VARCHAR": 1043,
    "INTEGER": 23,
    "FLOAT": 700,
    "TIMESTAMP": 1114,
    "DATE": 1082,
  }
  return type_to_oid.get(value.upper()) or 1043


import psycopg2
from psycopg2.extras import RealDictCursor
import json
import redis
from quillsql.error import PgQueryError
from quillsql.db.db_helper import connect_to_db, run_query_by_db

## The TTL for new cache entries (default: 1h)
DEFAULT_CACHE_TTL = 24 * 60 * 60


class CachedConnection:
    def __init__(self, database_type, config, cache_config, using_connection_string):
        self.database_type = database_type
        self.using_connection_string = using_connection_string
        self.config = config
        self.connection = connect_to_db(database_type, config, using_connection_string)
        self.cache = self.get_cache(cache_config)
        self.ttl = cache_config and cache_config.get("ttl") or DEFAULT_CACHE_TTL
        self.orgId = None

    def get_cache(self, cache_config):
        cache_type = cache_config and cache_config.get("cache_type")
        if cache_type and cache_type == "redis" or cache_type == "rediss":
            return redis.Redis(
                host=cache_config.get("host", "localhost"),
                port=cache_config.get("port", 6379),
                username=cache_config.get("username", "default"),
                password=cache_config.get("password"),
            )
        return None

    def exec_with_reconnect(self, sql):
        reconnect_count = 0
        while True:
            self.connection = connect_to_db(self.database_type, self.config, self.using_connection_string)
            try:
                return run_query_by_db(self.database_type, sql, self.connection)
            except psycopg2.Error as err:
                reconnect_count = reconnect_count + 1
                if reconnect_count < 10:
                    continue
                else:
                    raise PgQueryError(err, sql, err.diag.statement_position)

    def exec(self, sql):
        try:
            return self.exec_with_reconnect(sql)
        except psycopg2.Error as err:
            raise PgQueryError(err, sql, err.diag.statement_position)

    def query(self, sql):
        if not self.cache:
            return self.exec(sql)

        key = f"{self.orgId}:{self.sql}"
        cached_result = self.cache.get(key)
        if cached_result:
            cached = json.loads(cached_result)
            return {"rows": cached, "fields": cached['fields']}
        else:
            new_result = self.exec(sql)
            new_result_string = json.dumps(new_result)
            self.cache.set(key, new_result_string, "EX", DEFAULT_CACHE_TTL)
            return {"rows": new_result, "fields": new_result['fields']}
